"""
/***************************************************************************
 qgSurf - plugin for Quantum GIS

 Processing of geological planes and surfaces

                              -------------------
        begin                : 2011-12-21
        copyright            : (C) 2011-2023 by Mauro Alberti
        email                : alberti.m65@gmail.com

 ***************************************************************************/

# licensed under the terms of GNU GPL 3

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from math import atan

from qgis.PyQt.QtWidgets import *

try:
    from osgeo import ogr
except ImportError:
    import ogr

from gst.core.geometries.grids.rasters import *
from gst.io.rasters.gdal_io import *
from gst.qgis.canvas import *
from gst.qgis.orientations import *
from gst.qgis.points import *
from gst.qgis.project import *
from gst.qt.messages import *

from ..configurations.general_params import *

intersections_color = [
    "blue",
    "white",
    "red",
    "yellow",
    "orange",
    "brown",
    "green",
    "pink",
    "darkblue",
    "gray"
]

class DemPlaneIntersectionWidget(QWidget):
    """
    Constructor
    
    """

    line_colors = [
        "white",
        "red",
        "blue",
        "yellow",
        "orange",
        "brown"
    ]

    required_text = '--  required  --'

    def __init__(self,
                 plugin_folder: str,
                 module_name: str,
                 canvas,
                 plugin_qaction):

        super(DemPlaneIntersectionWidget, self).__init__()

        self.plugin_folder = plugin_folder
        self.module_name = module_name

        self.canvas = canvas
        self.plugin = plugin_qaction

        self.setup_gui()
        self.initialize_parameters()

    def initialize_parameters(self):

        self.previous_tool = None
        self.use_elevation_from_dem = False
        self.map_tool = None

        self.reset_all_parameters_post_dem_change()

    def reset_all_parameters_post_dem_change(self):

        self.disable_canvas_tools()
        self.reset_all_io_parameters()

    def reset_source_point_parameters(self):

        self.reset_source_point_in_map_QPushButton.setEnabled(False)
        self.reset_source_point_and_intersections()

    def reset_all_io_parameters(self):

        self.reset_all_input_parameters()
        self.reset_intersections()

    def reset_all_input_parameters(self):

        self.reset_input_dem_parameters()
        self.reset_source_point()
        self.reset_input_geoplane_parameters()

    def reset_input_dem_parameters(self):

        self.dem = None
        self.grid = None
        self.current_source_point_z_value = None

    def disable_canvas_tools(self):

        self.set_source_point_in_map_QPushButton.setEnabled(False)
        self.reset_source_point_in_map_QPushButton.setEnabled(False)

        try:
            self.map_tool.canvasClicked.disconnect(self.update_source_point_location)
        except:
            pass

        try:
            self.restore_previous_map_tool(self.map_tool)
        except:
            pass

    def reset_source_point_and_intersections(self):

        self.reset_intersections()
        self.reset_source_point()

    def reset_intersections(self):

        if hasattr(self, 'intersections_markers_list'):
            self.remove_intersection_markers_from_canvas()
        self.reset_intersections_markers_and_values()

    def reset_intersections_markers_list(self):

        self.intersections_markers_list = []

    def reset_intersections_values(self):

        self.intersections_x = []
        self.intersections_y = []
        self.intersections_xprt = {}
        self.inters = None

    def reset_intersections_markers_and_values(self):

        self.reset_intersections_markers_list()
        self.reset_intersections_values()

    def reset_source_point(self):

        self.clear_source_point_qlineedit_forms()
        if hasattr(self, 'source_point_marker_list'):
            self.remove_source_point_marker_from_canvas()
        self.reset_source_point_coordinates()
        self.reset_source_point_marker_list()

    def reset_source_point_coordinates(self):

        self.source_point_x = None
        self.source_point_y = None
        self.source_point_z = None

    def reset_source_point_marker_list(self):

        self.source_point_marker_list = []

    def remove_intersection_markers_from_canvas(self):

        for mrk in self.intersections_markers_list:
            self.canvas.scene().removeItem(mrk)

    def remove_source_point_marker_from_canvas(self):

        for mrk in self.source_point_marker_list:
            self.canvas.scene().removeItem(mrk)

    def reset_input_geoplane_parameters(self):

        self.source_dip_direction = None
        self.source_dip_angle = None

    def refresh_raster_layer_list(self):

        try:
            self.choose_source_dem_QComboBox.currentIndexChanged[int].disconnect(self.get_chosen_dem)
        except:
            pass

        try:
            self.reset_all_parameters_post_dem_change()
        except:
            pass

        try:
            self.choose_source_dem_QComboBox.clear()
        except:
            return

        self.choose_source_dem_QComboBox.addItem(self.required_text)

        self.raster_layers = loaded_raster_layers()
        if self.raster_layers is None or len(self.raster_layers) == 0:
            return
        for layer in self.raster_layers:
            self.choose_source_dem_QComboBox.addItem(layer.name())

        self.choose_source_dem_QComboBox.currentIndexChanged[int].connect(self.get_chosen_dem)

    def setup_gui(self):

        layout = QVBoxLayout()

        widget = QTabWidget()

        widget.addTab(self.setup_point_intersections_tab(), "Processing")
        widget.addTab(self.setup_help_tab(), "Help")
        
        layout.addWidget(widget)

        self.setLayout(layout)
        self.adjustSize()

        self.setWindowTitle(f'{plugin_name} - {self.module_name}')

    def setup_point_intersections_tab(self):
        
        widget = QWidget()

        layout = QVBoxLayout()
        layout.addWidget(self.setup_source_dem())
        layout.addWidget(self.setup_tabs())

        widget.setLayout(layout)
        return widget

    def setup_source_dem(self):

        widget = QWidget()

        layout = QGridLayout()

        self.refresh_raster_layer_list_QPushButton = QPushButton("Update raster layer list")
        self.refresh_raster_layer_list_QPushButton.clicked[bool].connect(self.refresh_raster_layer_list)
        self.refresh_raster_layer_list_QPushButton.setEnabled(True)
        layout.addWidget(self.refresh_raster_layer_list_QPushButton, 0, 0, 1, 3)

        layout.addWidget(QLabel("Choose current DEM layer"), 1, 0, 1, 1)
        self.choose_source_dem_QComboBox = QComboBox()
        self.choose_source_dem_QComboBox.addItem(self.required_text)
        layout.addWidget(self.choose_source_dem_QComboBox, 1, 1, 1, 2)

        self.refresh_raster_layer_list()

        widget.setLayout(layout)
        return widget

    def setup_tabs(self):  

        widget = QWidget()

        layout = QVBoxLayout()

        toolbox = QToolBox()
        toolbox.addItem(self.setup_geographicdata_sect(), 'Geographic parameters')
        toolbox.addItem(self.setup_geologicdata_sect(), 'Geological parameters')
        toolbox.addItem(self.setup_calculations_sect(), 'Calculate')
        toolbox.addItem(self.setup_output_section(), 'Output')

        layout.addWidget(toolbox)
        widget.setLayout(layout)

        return widget

    def setup_help_tab(self):
        
        widget = QWidget()
        layout = QVBoxLayout()

        # About section

        text_browser = QTextBrowser(widget)

        url_path = f"file:///{self.plugin_folder}/help/help_di.html"
        text_browser.setSource(QUrl(url_path))
        text_browser.setSearchPaths([f'{self.plugin_folder}/help/images/di'])

        layout.addWidget(text_browser)

        widget.setLayout(layout)

        return widget

    def setup_geographicdata_sect(self):        
        
        widget = QWidget()

        layout = QGridLayout()

        self.set_source_point_in_map_QPushButton = QPushButton("Set source point in map")
        self.set_source_point_in_map_QPushButton.clicked[bool].connect(self.set_source_point_in_map)
        layout.addWidget(self.set_source_point_in_map_QPushButton, 0, 0, 1, 3)

        self.reset_source_point_in_map_QPushButton = QPushButton("Reset source point")
        self.reset_source_point_in_map_QPushButton.clicked[bool].connect(self.reset_source_point_parameters)
        layout.addWidget(self.reset_source_point_in_map_QPushButton, 1, 0, 1, 3)
                      
        layout.addWidget(QLabel("X"), 2, 0, 1, 1)
        self.source_point_x_QLineEdit = QLineEdit()
        self.source_point_x_QLineEdit.textEdited.connect(self.update_source_point_location)
        layout.addWidget(self.source_point_x_QLineEdit, 2, 1, 1, 2)
              
        layout.addWidget(QLabel("Y"), 3, 0, 1, 1)
        self.source_point_y_QLineEdit = QLineEdit()
        self.source_point_y_QLineEdit.textEdited.connect(self.update_source_point_location)
        layout.addWidget(self.source_point_y_QLineEdit, 3, 1, 1, 2)
                      
        layout.addWidget(QLabel("Z"), 4, 0, 1, 1)
        self.source_point_z_QLineEdit = QLineEdit()
        self.source_point_z_QLineEdit.textEdited.connect(self.check_z_congruence_with_dem)
        layout.addWidget(self.source_point_z_QLineEdit, 4, 1, 1, 2)
               
        self.lock_z_value_to_dem_surface_QCheckBox = QCheckBox("lock z value to DEM surface")
        self.lock_z_value_to_dem_surface_QCheckBox.setChecked(True)
        self.lock_z_value_to_dem_surface_QCheckBox.stateChanged[int].connect(self.update_z_value)
        layout.addWidget(self.lock_z_value_to_dem_surface_QCheckBox, 5, 0, 1, 3)

        widget.setLayout(layout)
        
        return widget

    def update_crs_settings(self):

        self.map_crs = self.canvas.mapSettings().destinationCrs()

    def check_z_congruence_with_dem(self):
        
        if self.use_elevation_from_dem and float(self.source_point_z_QLineEdit.text()) != self.current_source_point_z_value:
            self.use_elevation_from_dem = False
            self.lock_z_value_to_dem_surface_QCheckBox.setChecked(False)
            
        self.current_source_point_z_value = float(self.source_point_z_QLineEdit.text())

    def setup_geologicdata_sect(self):

        widget = QWidget()
        layout = QGridLayout()
        
        dip_dir_label = QLabel("Dip direction")
        dip_dir_label.setAlignment (Qt.AlignCenter)       
        layout.addWidget(dip_dir_label, 0, 0, 1, 2)

        dip_ang_label = QLabel("Dip angle")
        dip_ang_label.setAlignment(Qt.AlignCenter)       
        layout.addWidget(dip_ang_label, 0, 2, 1, 1)
        
        self.dip_direction_QDial = QDial()
        self.dip_direction_QDial.setRange(0, 360)
        self.dip_direction_QDial.setPageStep(1)
        self.dip_direction_QDial.setProperty("value", 180)
        self.dip_direction_QDial.setSliderPosition(180)
        self.dip_direction_QDial.setTracking(True)
        self.dip_direction_QDial.setOrientation(Qt.Vertical)
        self.dip_direction_QDial.setWrapping(True)
        self.dip_direction_QDial.setNotchTarget(30.0)
        self.dip_direction_QDial.setNotchesVisible(True)
        self.dip_direction_QDial.valueChanged[int].connect(self.update_dip_direction_spinbox)
        layout.addWidget(self.dip_direction_QDial, 1, 0, 1, 2)
                
        self.DAngle_verticalSlider = QSlider()
        self.DAngle_verticalSlider.setRange(0, 90)
        self.DAngle_verticalSlider.setProperty("value", 45)
        self.DAngle_verticalSlider.setOrientation(Qt.Vertical)
        self.DAngle_verticalSlider.setInvertedAppearance(True)
        self.DAngle_verticalSlider.setTickPosition(QSlider.TicksBelow)
        self.DAngle_verticalSlider.setTickInterval(15)
        self.DAngle_verticalSlider.valueChanged[int].connect(self.update_dip_angle_spinbox)
        layout.addWidget(self.DAngle_verticalSlider, 1, 2, 1, 1)

        self.dip_direction_QDoubleSpinBox = QDoubleSpinBox()
        self.dip_direction_QDoubleSpinBox.setRange(0.0, 359.999)
        self.dip_direction_QDoubleSpinBox.setDecimals(3)
        self.dip_direction_QDoubleSpinBox.setSingleStep(1.0)
        self.dip_direction_QDoubleSpinBox.valueChanged[float].connect(self.update_dip_direction_slider)
        layout.addWidget(self.dip_direction_QDoubleSpinBox, 2, 0, 1, 2)
         
        self.dip_angle_QDoubleSpinBox = QDoubleSpinBox()
        self.dip_angle_QDoubleSpinBox.setRange(0.0, 90.0)
        self.dip_angle_QDoubleSpinBox.setDecimals(3)
        self.dip_angle_QDoubleSpinBox.setSingleStep(0.1)
        self.dip_angle_QDoubleSpinBox.setProperty("value", 45)
        self.dip_angle_QDoubleSpinBox.valueChanged[float].connect(self.update_dip_angle_slider)
        layout.addWidget(self.dip_angle_QDoubleSpinBox, 2, 2, 1, 1)

        widget.setLayout(layout)

        return widget
        
    def setup_calculations_sect(self):

        widget = QWidget()
        layout = QVBoxLayout()

        dem_groupbox = QGroupBox(self.tr("Intersections of DEM with geological plane -> intersection points"))

        dem_layout = QGridLayout()

        dem_layout.addWidget(QLabel("Marker color:"), 1, 0, 1, 1)

        self.intersections_color_QComboBox = QComboBox()
        self.intersections_color_QComboBox.insertItems(
            0,
            intersections_color
        )

        self.intersections_color_QComboBox.currentIndexChanged[int].connect(self.plot_intersections)

        dem_layout.addWidget(self.intersections_color_QComboBox, 1, 1, 1, 1)

        self.calculate_intersections_QPushButton = QPushButton("Calculate")
        self.calculate_intersections_QPushButton.clicked[bool].connect(self.calculate_intersections)
        dem_layout.addWidget(self.calculate_intersections_QPushButton, 2, 0, 1, 4)

        self.reset_intersections_QPushButton = QPushButton("Reset")
        self.reset_intersections_QPushButton.clicked.connect(self.reset_intersections)
        dem_layout.addWidget(self.reset_intersections_QPushButton, 3, 0, 1, 4)

        dem_groupbox.setLayout(dem_layout)

        layout.addWidget(dem_groupbox)

        # Distances of DEM points from geological plane

        distances_groupbox = QGroupBox(self.tr("Distances of DEM cell centers from geological plane -> grid of distances"))

        distances_layout = QGridLayout()

        distances_layout.addWidget(QLabel("Minimum value under plane"), 0, 0, 1, 1)

        self.lower_threshold_QDoubleSpinBox = QDoubleSpinBox()
        self.lower_threshold_QDoubleSpinBox.setRange(-1000.0, 1000.0)
        self.lower_threshold_QDoubleSpinBox.setValue(-50.0)
        distances_layout.addWidget(self.lower_threshold_QDoubleSpinBox, 0, 1, 1, 1)

        distances_layout.addWidget(QLabel("Maximum value above plane"), 0, 2, 1, 1)

        self.above_thresold_QDoubleSpinBox = QDoubleSpinBox()
        self.lower_threshold_QDoubleSpinBox.setRange(-1000.0, 1000.0)
        self.lower_threshold_QDoubleSpinBox.setValue(50.0)
        distances_layout.addWidget(self.above_thresold_QDoubleSpinBox, 0, 3, 1, 1)

        distances_layout.addWidget(QLabel(self.tr("Save results in")), 1, 0, 1, 1)

        self.output_grid_path_input_QLineEdit = QLineEdit()
        distances_layout.addWidget(self.output_grid_path_input_QLineEdit, 1, 1, 1, 2)

        output_browser = QPushButton(".....")
        output_browser.clicked.connect(self.select_output_grid_file)
        distances_layout.addWidget(output_browser, 1, 3, 1, 1)

        self.load_output_grid_in_map_QCheckBox = QCheckBox("load output grid in map")
        distances_layout.addWidget(self.load_output_grid_in_map_QCheckBox, 2, 0, 1, 2)

        self.calculate_distance_grid_QPushButton = QPushButton("Calculate distances grid")
        self.calculate_distance_grid_QPushButton.clicked[bool].connect(self.calculate_distances_grid)
        distances_layout.addWidget(self.calculate_distance_grid_QPushButton, 2, 2, 1, 2)

        distances_groupbox.setLayout(distances_layout)

        layout.addWidget(distances_groupbox)

        # final setup

        widget.setLayout(layout)

        return widget

    def setup_output_section(self):

        widget = QWidget()

        layout = QGridLayout()

        layout.addWidget(QLabel(self.tr("Save results in")), 0, 0, 1, 1)

        self.output_filename_input_QLineEdit = QLineEdit()
        layout.addWidget(self.output_filename_input_QLineEdit, 0, 1, 1, 2)

        self.output_vector_path_QPushButton = QPushButton(".....")
        self.output_vector_path_QPushButton.clicked.connect(self.select_output_vector_file_path)
        layout.addWidget(self.output_vector_path_QPushButton, 0, 3, 1, 1)

        layout.addWidget(QLabel(self.tr("with geometry:")), 1, 0, 1, 1)

        save_qbuttongroup = QButtonGroup(widget)

        self.save_as_point_shapefile_QRadioButton = QRadioButton("points")
        self.save_as_point_shapefile_QRadioButton.setChecked(True)
        save_qbuttongroup.addButton(self.save_as_point_shapefile_QRadioButton, 0)
        layout.addWidget(self.save_as_point_shapefile_QRadioButton, 1, 1, 1, 1)

        self.save_as_line_shapefile_QRadioButton = QRadioButton("lines")
        self.save_as_line_shapefile_QRadioButton.setChecked(True)
        save_qbuttongroup.addButton(self.save_as_line_shapefile_QRadioButton, 0)
        layout.addWidget(self.save_as_line_shapefile_QRadioButton, 1, 2, 1, 1)

        self.load_output_in_map_QCheckBox = QCheckBox("load output in project")
        layout.addWidget(self.load_output_in_map_QCheckBox, 2, 0, 1, 2)

        self.export_last_intersection_result_QPushButton = QPushButton("Save last intersections")
        self.export_last_intersection_result_QPushButton.clicked.connect(self.export_last_intersection_result)
        layout.addWidget(self.export_last_intersection_result_QPushButton, 3, 0, 1, 4)

        widget.setLayout(layout)

        return widget

    def set_source_point_in_map(self):
        
        try:
            self.map_tool.canvasClicked.disconnect(self.update_source_point_location)
        except:
            pass            
         
        self.update_crs_settings()
                       
        self.map_tool = EmitPointFromCanvas(self.canvas, self.plugin)  # mouse listener
        self.previous_tool = self.canvas.mapTool()  # save the standard map tool for restoring it at the end
        self.map_tool.canvasClicked.connect(self.update_source_point_location)
        self.map_tool.setCursor(Qt.CrossCursor)
        self.canvas.setMapTool(self.map_tool)

    def get_chosen_dem(self, ndx_dem_file = 0):

        self.dem = None        
        self.reset_all_parameters_post_dem_change()
        if self.raster_layers is None or len(self.raster_layers) == 0:
            return          
                                
        # no DEM layer defined  
        if ndx_dem_file == 0:
            return             

        self.dem = self.raster_layers[ndx_dem_file - 1]

        dem_src = self.dem.source()

        results, err = read_raster_band_with_projection(
            raster_source=dem_src)

        if err:
            QMessageBox.critical(
                self,
                "DEM",
                f"{repr(err)}"
            )
            return

        geotransform, epsg_code, band_params, data, dem_projection = results

        self.grid = Grid(
            array=data,
            geotransform=geotransform,
            epsg_code=epsg_code,
            projection=dem_projection
        )

        self.set_source_point_in_map_QPushButton.setEnabled(True)
        self.reset_source_point_in_map_QPushButton.setEnabled(True)

    def create_marker(
        self,
            canvas,
            prj_crs_x,
            prj_crs_y,
            pen_width=2,
            icon_type=2,
            icon_size=18,
            icon_color: Union[str, QColor] ='limegreen'
    ) -> QgsVertexMarker:
        
        marker = QgsVertexMarker(canvas)
        marker.setIconType(icon_type)
        marker.setIconSize(icon_size)
        marker.setPenWidth(pen_width)
        if isinstance(icon_color, QColor):
            color = icon_color
        else:
            color = QColor(icon_color)
        marker.setColor(color)
        marker.setCenter(QgsPointXY(prj_crs_x, prj_crs_y))

        return marker        

    def update_source_point_location(
            self,
            qgs_point=None,
    ):

        if self.grid is None:
            QMessageBox.critical(
                self,
                "Intersection source point",
                "Source DEM is not defined"
            )
            return

        self.reset_source_point_in_map_QPushButton.setEnabled(True)
               
        if self.sender() == self.map_tool:
            
            self.source_point_x = qgs_point.x()
            self.source_point_y = qgs_point.y()
            self.source_point_x_QLineEdit.setText(str(self.source_point_x))
            self.source_point_y_QLineEdit.setText(str(self.source_point_y))
            
        elif self.sender() == self.source_point_x_QLineEdit:
            
            if self.source_point_x_QLineEdit.text() == '':
                QMessageBox.critical(
                    self,
                    "qgSurf",
                    "Error: x value is not defined"
                )
                return
            try:
                self.source_point_x = float(self.source_point_x_QLineEdit.text())
            except:
                QMessageBox.critical(
                    self,
                    "qgSurf",
                    "Error: x value is not correctly defined"
                )
                return 
                                   
        elif self.sender() == self.source_point_y_QLineEdit:
            
            if self.source_point_y_QLineEdit.text() == '':
                QMessageBox.critical(
                    self,
                    "qgSurf",
                    "Error: y value is not defined")
                return  
            try:
                self.source_point_y = float(self.source_point_y_QLineEdit.text())
            except:
                QMessageBox.critical(
                    self,
                    "qgSurf",
                    "Error: y value is not correctly defined")
                return            
        
        z_value_from_dem = self.update_z_value()
        if z_value_from_dem is None: 
            self.current_source_point_z_value = None
            self.source_point_z_QLineEdit.setText("")
        else:
            self.current_source_point_z_value = z_value_from_dem
            self.source_point_z_QLineEdit.setText(str(self.current_source_point_z_value))
            self.use_elevation_from_dem = True
                           
        self.remove_intersection_markers_from_canvas()
        self.remove_source_point_marker_from_canvas()
        color = QColor(str(self.intersections_color_QComboBox.currentText()))

        source_point_marker = self.create_marker(
            self.canvas,
            self.source_point_x, self.source_point_y,
            icon_type=1,
            icon_color=color)

        self.source_point_marker_list = [source_point_marker]

        self.canvas.refresh()

    def update_z_value (self):
        """
        Update z value.
        
        """

        # does nothing when the height source is not from DEM

        if not self.lock_z_value_to_dem_surface_QCheckBox.isChecked():
            return None

        # prevent action when the DEM is not read

        if self.grid is None:
            return None 

        # get z value from ptLyr

        srcpt_demcrs_x, srcpt_demcrs_y = self.project_from_map_to_dem_crs(
            self.source_point_x,
            self.source_point_y
        )

        # return None or a numeric value
        return self.grid.interpolate_bilinear(
            srcpt_demcrs_x,
            srcpt_demcrs_y)

    def restore_previous_map_tool(self,
                                  map_tool):

        if map_tool:
            self.canvas.unsetMapTool(map_tool)

        if self.previous_tool:
            self.canvas.setMapTool(self.previous_tool)

    def clear_source_point_qlineedit_forms(self):
        
        for qlineedit_form in (
                self.source_point_x_QLineEdit,
                self.source_point_y_QLineEdit,
                self.source_point_z_QLineEdit):

            qlineedit_form.clear()

    def update_dip_direction_slider(self):
        """
        Update the value of the dip direction in the slider.""
        """        
        transformed_dipdirection = self.dip_direction_QDoubleSpinBox.value() + 180.0
        if transformed_dipdirection > 360.0:
            transformed_dipdirection -= 360.0
        self.dip_direction_QDial.setValue(int(transformed_dipdirection))

    def update_dip_direction_spinbox(self):
        """
        Update the value of the dip direction in the spinbox.
        """        
        real_dipdirection = self.dip_direction_QDial.value() - 180.0
        if real_dipdirection < 0.0:
            real_dipdirection += 360.0
        self.dip_direction_QDoubleSpinBox.setValue(real_dipdirection)

    def update_dip_angle_slider(self):
        """
        Update the value of the dip angle in the slider.
        """
        
        self.DAngle_verticalSlider.setValue(int(self.dip_angle_QDoubleSpinBox.value()))

    def update_dip_angle_spinbox(self):
        """
        Update the value of the dip angle in the spinbox.
        """        
        
        self.dip_angle_QDoubleSpinBox.setValue(self.DAngle_verticalSlider.value())

    def project_from_map_to_dem_crs(
        self,
        x: float,
        y: float
    ) -> Tuple[float, float]:
        """
        Converts point coordinates from project to DEM crs.

        :param x: x coordinate in the project CRS.
        :param y: y coordinate in the project CRS.
        :return: the point coordinates in the DEM crs.
        """

        return qgs_project_xy(
            x=x,
            y=y,
            src_crs=self.map_crs,
            dest_crs=self.dem.crs()
        )

    def project_from_dem_to_map_crs(
        self,
        x: float,
        y: float
    ) -> Tuple[float, float]:
        """
        Converts point coordinates from project to DEM crs.

        :param x: x coordinate in the DEM CRS.
        :param y: y coordinate in the DEM CRS.
        :return: the point coordinates in the project crs.
        """

        return qgs_project_xy(
            x,
            y,
            self.dem.crs(),
            self.map_crs
        )

    def parse_input_parameters(self
                               ) -> Tuple[Union[type(None), numbers.Real], Error]:

        try:

            # check if all input data are correct

            if self.grid is None:
                z = None
                msg = "Please first define a source DEM"
            elif self.source_point_x_QLineEdit.text() == '' or self.source_point_y_QLineEdit.text() == '' or self.source_point_z_QLineEdit.text() == '':
                z = None
                msg = "Define the location of the source point in 'Geographic parameters' section"
            else:
                try:
                    z = float(self.source_point_z_QLineEdit.text())
                    msg = ''
                except:
                    z = None
                    msg = "z value is not correctly defined"

            if z is None:
                err = Error(
                    True,
                    caller_name(),
                    Exception(msg)
                )
            else:
                err = Error()

            return z, err

        except Exception as e:

            return None, Error(
                True,
                caller_name(),
                e,
                traceback.format_exc()
            )

    def preprocess_plane_and_source_point(
        self,
        prj_crs_src_pt_x,
        prj_crs_src_pt_y
    ) -> Tuple[Union[type(None), Tuple[Plane, Point]], Error]:

        try:

            self.source_dip_direction = self.dip_direction_QDoubleSpinBox.value()
            self.source_dip_angle = self.dip_angle_QDoubleSpinBox.value()

            # Calculates dip direction correction with respect to map CRS top (y-axis) direction

            geographic_north_direction_azimuth = calculate_local_geographic_north_azimuth(
                crs=self.map_crs,
                src_pt=Point(prj_crs_src_pt_x, prj_crs_src_pt_y),
                )

            print(f"Azimuth correction: {geographic_north_direction_azimuth}")

            dip_direction_in_map_crs = (self.source_dip_direction + geographic_north_direction_azimuth) % 360.0
            dip_angle_in_map_crs = self.source_dip_angle

            print(f"Project CRS corrected plane orientation: dip dir. = {dip_direction_in_map_crs}, dip ang. = {dip_angle_in_map_crs}")

            # Correct dip direction and angle by DEM crs

            prjcrs_dir_versor = Direct(
                az=dip_direction_in_map_crs,
                pl=0.0).as_versor()

            dummy_distance = 100  # meters

            prjcrs_dir_vector = prjcrs_dir_versor.scale(dummy_distance)

            # NOTE: it assumes that the project crs is not in lon-lat !!

            prjcrs_srcpt = Point(
                prj_crs_src_pt_x,
                prj_crs_src_pt_y)

            endpt_prjcrs_x, endpt_prjcrs_y = prjcrs_srcpt.shift2d_by_vect(prjcrs_dir_vector).xy()

            prjcrs_zeta = dummy_distance * tan(radians(dip_angle_in_map_crs))

            srcpt_demcrs_x, srcpt_demcrs_y = self.project_from_map_to_dem_crs(
                x=prj_crs_src_pt_x,
                y=prj_crs_src_pt_y)

            dem_crs_source_point = Point(srcpt_demcrs_x, srcpt_demcrs_y)

            endpt_demcrs_x, endpt_demcrs_y = self.project_from_map_to_dem_crs(
                x=endpt_prjcrs_x,
                y=endpt_prjcrs_y
            )

            # geoplane attitude in DEM CRS

            dem_crs_vector = Segment(
                start_pt=dem_crs_source_point,
                end_pt=Point(endpt_demcrs_x, endpt_demcrs_y)
            ).as_vector2d()

            corr_dip_direction = dem_crs_vector.azimuth_degr()

            demcrs_len = dem_crs_vector.length
            demcrs_zeta = prjcrs_zeta

            corr_dip_angle = degrees(atan(demcrs_zeta / demcrs_len))

            srcpt_3d_demcrs = Point(
                srcpt_demcrs_x,
                srcpt_demcrs_y,
                self.source_point_z
            )

            return (
                        Plane(
                            corr_dip_direction,
                            corr_dip_angle
                        ),
                        srcpt_3d_demcrs
                    ), Error()

        except Exception as e:

            return None, Error(
                True,
                caller_name(),
                e,
                traceback.format_exc()
            )

    def calculate_intersections(self):
        """
        Calculate intersection points.
        """

        try:

            result, err = self.parse_input_parameters()

            if err:
                QMessageBox.critical(self, "qgSurf", f"{err}")
                return

            self.source_point_z = result

            # source point

            result, err = self.preprocess_plane_and_source_point(
                self.source_point_x,
                self.source_point_y
            )

            if err:
                QMessageBox.critical(self, "qgSurf", f"{repr(err)}")
                return

            self.geological_plane_in_dem_crs, source_point_in_dem_crs = result

            result = plane_dem_intersections_strategy_1(
                plane_attitude=self.geological_plane_in_dem_crs,
                source_point=source_point_in_dem_crs,
                grid=self.grid
            )

            if isinstance(result, Error):
                err = result
                QMessageBox.critical(self, "qgSurf", f"{repr(err)}")
                return

            xcoords_x, xcoords_y, ycoords_x, ycoords_y = result

            self.inters = Intersections()

            self.inters.xcoords_x = xcoords_x
            self.inters.xcoords_y = xcoords_y
            self.inters.ycoords_x = ycoords_x
            self.inters.ycoords_y = ycoords_y

            self.inters.parameters = Intersection_Parameters(
                source_point_in_dem_crs,
                self.geological_plane_in_dem_crs
            )

            self.intersections_x = list(xcoords_x[np.logical_not(np.isnan(xcoords_x))]) + \
                                     list(ycoords_x[np.logical_not(np.isnan(ycoords_y))])

            self.intersections_y = list(xcoords_y[np.logical_not(np.isnan(xcoords_x))]) + \
                                     list(ycoords_y[np.logical_not(np.isnan(ycoords_y))])

            intersection_data = dict(
                x=self.intersections_x,
                y=self.intersections_y)

            intersection_plane = dict(
                dipdir=self.inters.parameters._srcPlaneAttitude._dipdir,
                dipangle=self.inters.parameters._srcPlaneAttitude._dipangle
            )

            intersection_point = dict(
                x=self.inters.parameters._srcPt.x,
                y=self.inters.parameters._srcPt.y,
                z=self.inters.parameters._srcPt.z
            )

            self.intersections_xprt = dict(
                data=intersection_data,
                plane=intersection_plane,
                point=intersection_point)

            self.plot_intersections()

        except Exception as e:

            QMessageBox.critical(self, "qgSurf", f"Exception: {repr(e)}")
            return

    def calculate_distances_grid(self):
        """
        Calculate distances of DEM cell centers from geological plane.
        """

        try:

            output_grid_path = self.output_grid_path_input_QLineEdit.text()

            result, err = self.parse_input_parameters()

            if err:
                QMessageBox.information(
                    self,
                    f"{plugin_name} - {self.module_name}",
                    f"{repr(err)}")
                return

            self.source_point_z = result

            # source point

            result, err = self.preprocess_plane_and_source_point(
                self.source_point_x,
                self.source_point_y
            )

            if err:
                QMessageBox.critical(self, "qgSurf", f"{repr(err)}")
                return

            self.geological_plane_in_dem_crs, source_point_in_dem_crrs = result

            # define distance array limits

            below_threshold = float(self.lower_threshold_QDoubleSpinBox.value())
            above_threshold = float(self.upper_threshold_spinbox.value())

            if below_threshold >= above_threshold:
                QMessageBox.critical(self, "qgSurf", f"Below threshol {below_threshold} is not lower than upper threshold {above_threshold} ")
                return

            # populate array with distances from DEM

            cplane = CPlane3D.from_geol_plane(
                geol_plane=self.geological_plane_in_dem_crs,
                pt=source_point_in_dem_crrs
            )

            arr = self.grid.array
            num_rows, num_cols = arr.shape
            distances = np.zeros((num_rows, num_cols)) * np.nan

            for i in range(num_rows):
                for j in range(num_cols):
                    val = arr[i, j]
                    if not np.isfinite(val):
                        continue
                    cc_pt_x, cc_pt_y = self.grid.ijArrToxy(i, j)
                    dem_cell_center = Point(
                        cc_pt_x,
                        cc_pt_y,
                        val
                    )

                    distance = cplane.signed_distance_to_point(dem_cell_center)
                    if below_threshold <= distance <= above_threshold:
                        distances[i, j] = distance

            err = write_geotiff(
                file_path=output_grid_path,
                arr_out=distances,
                geotransform=GeoTransform.to_gdal_geotransform(self.grid.geotransform),
                projection=self.grid.projection
            )

            if err:
                QMessageBox.critical(self, "qgSurf", f"{repr(err)}")
                return

            if self.load_output_grid_in_map_QCheckBox.isChecked():

                try:

                    distances_grid_layer = QgsRasterLayer(
                        output_grid_path,
                        QFileInfo(output_grid_path).baseName()
                    )
                    QgsProject.instance().addMapLayer(distances_grid_layer)

                    if distances_grid_layer.isValid():
                        print("Grid layer was loaded successfully")
                    else:
                        QMessageBox.critical(
                            self,
                            "qgSurf",
                            "Unable to read basename and file path - Output string is possibly invalid")

                except Exception as e:

                    QMessageBox.critical(
                        self,
                        "qgSurf",
                        f"Unable to load grid layer in project: {repr(e)}"
                    )
                    return

            QMessageBox.information(self, "qgSurf", f"Distance grid created")

        except Exception as e:

            QMessageBox.critical(self, "qgSurf", f"Exception: {repr(e)}")
            return

    def plot_intersections(self):

        try:

            current_markers_list = []

            for x, y in zip(self.intersections_x, self.intersections_y):
                prj_crs_x, prj_crs_y = self.project_from_dem_to_map_crs(x, y)
                marker = self.create_marker(
                    self.canvas,
                    prj_crs_x,
                    prj_crs_y,
                    pen_width=1,
                    icon_type=1,
                    icon_size=8,
                    icon_color=QColor(str(self.intersections_color_QComboBox.currentText())))

                current_markers_list.append(marker)

            self.intersections_markers_list += current_markers_list
            self.canvas.refresh()

        except Exception as e:

            return Error(
                True,
                caller_name(),
                e,
                traceback.format_exc())

    def select_output_vector_file_path(self):
            
        output_filename, __ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save shapefile"),
            "*.shp",
            "shp (*.shp *.SHP)"
        )

        if not output_filename: 
            return

        self.output_filename_input_QLineEdit.setText(output_filename)

    def select_output_grid_file(self):

        output_filename, __ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save grid as geotiff"),
            "*.tif",
            "tif (*.tif *.TIF)"
        )

        if not output_filename:
            return

        self.output_grid_path_input_QLineEdit.setText(output_filename)

    def export_last_intersection_result(self):
        """
        Write intersection results in the output shapefile.
        """

        # check for result existence

        if self.inters is None:
            QMessageBox.critical(
                self,
                "Save results",
                "No results available")
            return

        self.output_filename = str(self.output_filename_input_QLineEdit.text())
        if self.output_filename == '':
            QMessageBox.critical(
                self,
                "Save results",
                "No output file defined"
            )
            return

        # set output type
        if self.save_as_point_shapefile_QRadioButton.isChecked():
            self.result_geometry = 'points'
        else:
            self.result_geometry = 'lines'

        # creation of output shapefile
        shape_driver = ogr.GetDriverByName("ESRI Shapefile")
        self.out_shape = shape_driver.CreateDataSource(self.output_filename)
        if self.out_shape is None:
            QMessageBox.critical(self, "Results", "Unable to create output shapefile: %s" % self.output_filename)
            return

        if self.result_geometry == 'points':
            self.out_layer = self.out_shape.CreateLayer('output_lines', geom_type=ogr.wkbPoint)
        else:
            self.out_layer = self.out_shape.CreateLayer('output_lines', geom_type=ogr.wkbLineString)

        if self.out_layer is None:
            QMessageBox.critical(
                self,
                "Results",
                f"Unable to create output shapefile: {self.output_filename}"
            )
            return

        # set analysis parameters
        sourcePoint = self.inters.parameters._srcPt
        self.srcPlaneAttitude = self.inters.parameters._srcPlaneAttitude
        self.z_from_plane_closure = self.srcPlaneAttitude.plane_from_geo(sourcePoint)

        # add fields to the output shapefile
        id_fieldDef = ogr.FieldDefn('id', ogr.OFTInteger)
        self.out_layer.CreateField(id_fieldDef)

        if self.result_geometry == 'points':
            x_fieldDef = ogr.FieldDefn('x', ogr.OFTReal)
            self.out_layer.CreateField(x_fieldDef)
            y_fieldDef = ogr.FieldDefn('y', ogr.OFTReal)
            self.out_layer.CreateField(y_fieldDef)
            z_fieldDef = ogr.FieldDefn('z', ogr.OFTReal)
            self.out_layer.CreateField(z_fieldDef)

        srcPt_x_fieldDef = ogr.FieldDefn('srcPt_x', ogr.OFTReal)
        self.out_layer.CreateField(srcPt_x_fieldDef)
        srcPt_y_fieldDef = ogr.FieldDefn('srcPt_y', ogr.OFTReal)
        self.out_layer.CreateField(srcPt_y_fieldDef)
        srcPt_z_fieldDef = ogr.FieldDefn('srcPt_z', ogr.OFTReal)
        self.out_layer.CreateField(srcPt_z_fieldDef)
        DipDir_fieldDef = ogr.FieldDefn('dip_dir', ogr.OFTReal)
        self.out_layer.CreateField(DipDir_fieldDef)
        DipAng_fieldDef = ogr.FieldDefn('dip_ang', ogr.OFTReal)
        self.out_layer.CreateField(DipAng_fieldDef)

        # get the layer definition of the output shapefile
        self.outshape_featdef = self.out_layer.GetLayerDefn()

        # write results
        if self.result_geometry == 'points':
            self.write_intersections_as_points()
        else:
            self.write_intersections_as_lines()

        QMessageBox.information(self, "Result", "Saved to shapefile: %s" % self.output_filename)

        # add theme to QGis project
        if self.load_output_in_map_QCheckBox.isChecked():
            try:
                intersection_layer = QgsVectorLayer(
                    self.output_filename,
                    QFileInfo(self.output_filename).baseName(),
                    "ogr")

                QgsProject.instance().addMapLayer(intersection_layer)
            except:
                QMessageBox.critical(self, "Result", "Unable to load layer in project")
                return

    def write_intersections_as_points(self):
        """
        Write intersection results in the output shapefile.
        """

        x_filtered_coord_x = self.inters.xcoords_x[np.logical_not(np.isnan(self.inters.xcoords_x))]
        x_filtered_coord_y = self.inters.xcoords_y[np.logical_not(np.isnan(self.inters.xcoords_x))]
        x_filtered_coord_z = self.z_from_plane_closure(x_filtered_coord_x, x_filtered_coord_y)

        y_filtered_coord_x = self.inters.ycoords_x[np.logical_not(np.isnan(self.inters.ycoords_y))]
        y_filtered_coord_y = self.inters.ycoords_y[np.logical_not(np.isnan(self.inters.ycoords_y))]
        y_filtered_coord_z = self.z_from_plane_closure(y_filtered_coord_x, y_filtered_coord_y)

        intersections_x = list(x_filtered_coord_x) + list(y_filtered_coord_x)
        intersections_y = list(x_filtered_coord_y) + list(y_filtered_coord_y)
        intersections_z = list(x_filtered_coord_z) + list(y_filtered_coord_z)

        curr_Pt_id = 0
        for curr_Pt in zip(intersections_x, intersections_y, intersections_z):
            curr_Pt_id += 1
            prj_crs_x, prj_crs_y = self.project_from_dem_to_map_crs(float(curr_Pt[0]), float(curr_Pt[1]))

            # pre-processing for new feature in output layer
            curr_Pt_geom = ogr.Geometry(ogr.wkbPoint)
            curr_Pt_geom.AddPoint(prj_crs_x, prj_crs_y, float(curr_Pt[2]))

            # create a new feature
            curr_Pt_shape = ogr.Feature(self.outshape_featdef)
            curr_Pt_shape.SetGeometry(curr_Pt_geom)
            curr_Pt_shape.SetField('id', curr_Pt_id)

            curr_Pt_shape.SetField('x', prj_crs_x)
            curr_Pt_shape.SetField('y', prj_crs_y)
            curr_Pt_shape.SetField('z', curr_Pt[2])

            prj_crs_src_pt_x, prj_crs_src_pt_y = self.project_from_dem_to_map_crs(self.intersections_xprt['point']['x'],
                                                                                  self.intersections_xprt['point']['y'])

            curr_Pt_shape.SetField('srcPt_x', prj_crs_src_pt_x)
            curr_Pt_shape.SetField('srcPt_y', prj_crs_src_pt_y)
            curr_Pt_shape.SetField('srcPt_z', self.intersections_xprt['point']['z'])

            curr_Pt_shape.SetField('dip_dir', self.srcPlaneAttitude._dipdir)
            curr_Pt_shape.SetField('dip_ang', self.srcPlaneAttitude._dipangle)

            # add the feature to the output layer
            self.out_layer.CreateFeature(curr_Pt_shape)

            # destroy no longer used objects
            try:
                curr_Pt_geom.Destroy()
                curr_Pt_shape.Destroy()
            except:
                pass

        # destroy output geometry
        try:
            self.out_shape.Destroy()
        except:
            pass

    def write_intersections_as_lines(self):
        """
        Write intersection results in a line shapefile.
        """

        # create dictionary of cell with intersection points
        self.inters.links = self.inters.get_intersections()
        self.inters.neighbours = self.inters.set_neighbours()
        self.inters.define_paths()

        # networks of connected intersections
        self.inters.networks = self.inters.define_networks()

        for curr_path_id, curr_path_points in self.inters.networks.items():
            line = ogr.Geometry(ogr.wkbLineString)
            for curr_point_id in curr_path_points:
                curr_intersection = self.inters.links[curr_point_id - 1]
                i, j, direct = curr_intersection['i'], curr_intersection['j'], curr_intersection['pi_dir']
                if direct == 'x': dem_crs_x, dem_crs_y = self.inters.xcoords_x[i, j], self.inters.xcoords_y[i, j]
                if direct == 'y': dem_crs_x, dem_crs_y = self.inters.ycoords_x[i, j], self.inters.ycoords_y[i, j]
                z = self.z_from_plane_closure(dem_crs_x, dem_crs_y)
                prj_crs_x, prj_crs_y = self.project_from_dem_to_map_crs(dem_crs_x, dem_crs_y)
                line.AddPoint(prj_crs_x, prj_crs_y, z)

                # create a new feature
            line_shape = ogr.Feature(self.outshape_featdef)
            line_shape.SetGeometry(line)

            line_shape.SetField('id', curr_path_id)

            prj_crs_src_pt_x, prj_crs_src_pt_y = self.project_from_dem_to_map_crs(self.intersections_xprt['point']['x'],
                                                                                  self.intersections_xprt['point']['y'])

            line_shape.SetField('srcPt_x', prj_crs_src_pt_x)
            line_shape.SetField('srcPt_y', prj_crs_src_pt_y)
            line_shape.SetField('srcPt_z', self.intersections_xprt['point']['z'])

            line_shape.SetField('dip_dir', self.srcPlaneAttitude._dipdir)
            line_shape.SetField('dip_ang', self.srcPlaneAttitude._dipangle)

            # add the feature to the output layer
            self.out_layer.CreateFeature(line_shape)

            # destroy no longer used objects
            #line.Destroy()
            line_shape.Destroy()

        # destroy output geometry
        self.out_shape.Destroy()


