"""
/***************************************************************************
 qgSurf - plugin for Quantum GIS

 Processing of geological planes and surfaces

                              -------------------
        begin                : 2011-12-21
        copyright            : (C) 2011-2023 by Mauro Alberti
        email                : alberti.m65@gmail.com

 ***************************************************************************/

# licensed under the terms of GNU GPL 3

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from PyQt5.QtWidgets import *

from gst.core.geometries.planes import *
from gst.core.orientations.orientations import Plane as GPlane
from gst.io.vectors.ogr_io import *
from gst.qgis.canvas import *
from gst.qgis.messages import *
from gst.qgis.orientations import *
from gst.qgis.rasters import *
from gst.qgis.tables import *

from ..configurations.general_params import *
from ..configurations.plane_points_distances_params import *

from ..ui.auxiliary_windows.distances import DistancesSrcPtLyrDlg, undefined_text


def parse_field_choice(
    val,
    choose_message
):

    if val == choose_message:
        return None
    else:
        return val


def get_module_input_parameters(dialog):

    try:

        pt_layer = dialog.point_layer

        point_layer_id_name_field = parse_field_choice(
            dialog.cmbIdSrcFld.currentText(),
            undefined_text)
        if not point_layer_id_name_field:
            return False, "Incorrect id field choice"

        x_name_field = parse_field_choice(
            dialog.cmbXSrcFld.currentText(),
            undefined_text)

        if not x_name_field:
            return False, "Incorrect x field choice"

        y_name_field = parse_field_choice(
            dialog.cmbYSrcFld.currentText(),
            undefined_text)

        if not y_name_field:
            return False, "Incorrect y field choice"

        z_name_field = parse_field_choice(
            dialog.cmbZSrcFld.currentText(),
            undefined_text)

        if not z_name_field:
            return False, "Incorrect z field choice"

        # geological plane parameters section

        plane_dip_direction = dialog.spnTargetAttDipDir.value()
        plane_dip_angle = dialog.spnTargetAttDipAng.value()
        is_dip_direction_relative_to_geographic_north = dialog.DirectionAzimuthRelatedToGeographicNorth_QRadioButton.isChecked()

        try:

            x_coord_of_plane_source_point = float(dialog.SourcePointXFieldName_QLineEdit.text())
            y_coord_of_plane_source_point = float(dialog.SourcePointYFieldName_QLineEdit.text())
            z_coord_of_plane_source_point = float(dialog.SourcePointZFieldName_QLineEdit.text())

        except Exception as e:

            return False, f"Incorrect source point coordinates -> exception: {e!r}"

        # output shapefile

        output_shapefile_path = dialog.lnedtOutFilename.text()
        if not output_shapefile_path:
            return False, "Missing output shapefile path"

        # returning parameters

        return True, dict(
            point_layer=pt_layer,
            point_layer_flds=(point_layer_id_name_field, x_name_field, y_name_field, z_name_field),
            gplane_params=dict(
                pl_attitude=(plane_dip_direction, plane_dip_angle),
                src_pt=(x_coord_of_plane_source_point, y_coord_of_plane_source_point, z_coord_of_plane_source_point),
                is_dip_direction_relative_to_geographic_north=is_dip_direction_relative_to_geographic_north,),
            output_shapefile_path=output_shapefile_path
        )

    except Exception as e:

        return False, f"Error with input parameters: {e!r}"


class PointsPlaneDistancesWidget(QWidget):

    sgnWindowClosed = pyqtSignal()

    def __init__(self,
                 canvas,
                 plugin_name,
                 plugin_folder
                 ):

        super(PointsPlaneDistancesWidget, self).__init__()
        self.map_canvas = canvas
        self.plugin_name = plugin_name
        self.plugin_folder = plugin_folder

        self.point_layer = None
        self.point_layer_fields = []
        self.distances_analysis_params = None

        self.setup_gui()

    def setup_gui(self): 

        self.layout = QVBoxLayout()
        
        self.layout.addWidget(self.setup_input_data())
        self.layout.addWidget(self.setup_processing())
        self.layout.addWidget(self.setup_help())

        self.setLayout(self.layout)
        self.setWindowTitle(f"{self.plugin_name} - points distances to geological plane")
        self.adjustSize()

    def setup_input_data(self):
        
        grpInput = QGroupBox("Define params")
        
        layout = QVBoxLayout() 
        
        self.DefineIO_QPushButton = QPushButton(self.tr("I/O params"))
        self.DefineIO_QPushButton.clicked.connect(self.define_input_parameters)
        layout.addWidget(self.DefineIO_QPushButton)
        
        grpInput.setLayout(layout)
        
        return grpInput

    def setup_processing(self):
        
        grpProcessing = QGroupBox("Processing")
        
        layout = QGridLayout()

        self.CalculateDistances_QPushButton = QPushButton(self.tr("Calculate"))
        self.CalculateDistances_QPushButton.clicked.connect(self.calculate_distances_of_points_to_plane)
        layout.addWidget(self.CalculateDistances_QPushButton, 0, 0, 1, 1)

        grpProcessing.setLayout(layout)
        
        return grpProcessing

    def setup_help(self):

        group_box = QGroupBox("Help")

        layout = QVBoxLayout()

        self.OpenHelp_QPushButton = QPushButton(self.tr("Open help"))
        self.OpenHelp_QPushButton.clicked.connect(self.open_help)
        layout.addWidget(self.OpenHelp_QPushButton)

        group_box.setLayout(layout)

        return group_box

    def define_input_parameters(self):

        self.distances_analysis_params = None

        if len(loaded_point_layers()) == 0:
            warn_qgis(
                self.plugin_name,
                "No available point layers"
            )
            return

        dialog = DistancesSrcPtLyrDlg()
        if dialog.exec_():
            success, cntnt = get_module_input_parameters(dialog)
        else:
            warn_qgis(
                self.plugin_name,
                "No input data defined"
            )
            return

        if not success:
            warn_qgis(
                self.plugin_name,
                cntnt
            )
            return
        else:
            structural_input_params = cntnt

        self.point_layer = structural_input_params["point_layer"]
        self.point_layer_fields = structural_input_params["point_layer_flds"]
        self.geological_plane_attitude = structural_input_params["gplane_params"]["pl_attitude"]
        self.geological_plane_src_point = structural_input_params["gplane_params"]["src_pt"]
        self.is_dip_direction_relative_to_geographic_north = structural_input_params["gplane_params"]["is_dip_direction_relative_to_geographic_north"]
        self.output_shapefile_path = structural_input_params["output_shapefile_path"]

        info_qgis(
            self.plugin_name,
            f"Now results can be calculated"
        )

    def calculate_distances_of_points_to_plane(
        self,
    ):

        if not self.point_layer or \
           not self.point_layer_fields:
            warn_qgis(
                self.plugin_name,
                "Input data are not defined"
            )
            return

        # get input point layer data

        point_layer_data = extract_layer_attributes(
            layer=self.point_layer,
            fields=self.point_layer_fields
        )

        len_data = len(point_layer_data)

        if len_data == 0:
            warn_qgis(
                self.plugin_name,
                "No data to calculate"
            )
            return

        source_point_of_cartesian_plane = Point(*self.geological_plane_src_point)

        project_crs = extract_crs_of_project()

        plane_dip_direction, plane_dip_angle = self.geological_plane_attitude

        if self.is_dip_direction_relative_to_geographic_north:

            geographic_north_azimuth_degr = calculate_local_geographic_north_azimuth(
                crs=project_crs,
                src_pt=source_point_of_cartesian_plane,
            )

            print(f"Geographic North azimuth (degrees): {geographic_north_azimuth_degr}")

            plane_dip_direction = (plane_dip_direction + geographic_north_azimuth_degr) % 360.0

        cartesian_plane = CPlane3D.from_geol_plane(
            GPlane(
                plane_dip_direction,
                plane_dip_angle
            ),
            source_point_of_cartesian_plane
        )

        point_attributes = []
        geographic_point_coordinates = []

        for geog_pt_data in point_layer_data:

            pt_rec_id, pt_x, pt_y, pt_z = geog_pt_data
            geog_pt = Point(
                pt_x,
                pt_y,
                pt_z
            )

            prj_pt, err = cartesian_plane.project_point(geog_pt)
            if err:
                warn_qgis(
                    self.plugin_name,
                    f"{repr(err)}"
                )

            prj_x, prj_y, prj_z = prj_pt.xyz()

            geo_prj_ds = cartesian_plane.signed_distance_to_point(geog_pt)

            src_geog_vect = Segment(
                start_pt=source_point_of_cartesian_plane,
                end_pt=geog_pt).as_vector2d()

            src_prj_vect = Segment(
                start_pt=source_point_of_cartesian_plane,
                end_pt=prj_pt).as_vector2d()

            src_geo_ds = src_geog_vect.length
            src_prj_ds = src_prj_vect.length

            src_geo_az = src_geog_vect.azimuth_degr()
            src_prj_az = src_prj_vect.azimuth_degr()

            prj_geo_an = src_geog_vect.angle_with(src_prj_vect)

            vals = (
                pt_rec_id,
                pt_x,
                pt_y,
                pt_z,
                prj_x,
                prj_y,
                prj_z,
                geo_prj_ds,
                src_geo_ds,
                src_geo_az,
                src_prj_ds,
                src_prj_az,
                prj_geo_an
            )

            point_attributes.append(vals)
            geographic_point_coordinates.append((pt_x, pt_y, pt_z))

        field_names = [f["name"] for f in field_parameters_of_output_table]

        out_shape_pth = self.output_shapefile_path

        shapefile_datasource, point_shapelayer = shapefile_create(
            path=out_shape_pth,
            geom_type=ogr.wkbPoint25D,
            fields_dict_list=field_parameters_of_output_table
        )

        print(f"Shapefile {out_shape_pth} created")

        if not shapefile_datasource or not point_shapelayer:
            warn_qgis(
                self.plugin_name,
                f"Unable to create output shapefile {out_shape_pth}"
            )
            return

        success, msg = try_write_pt_shapefile(
            point_layer=point_shapelayer,
            geoms=geographic_point_coordinates,
            field_names=field_names,
            attrs=point_attributes)

        if success:
            info_qgis(
                self.plugin_name,
                f"Results written in output shapefile {out_shape_pth}"
            )
        else:
            warn_qgis(
                self.plugin_name,
                msg
            )

    def closeEvent(self, event):

        settings = QSettings(settings_name, plugin_name)

        settings.setValue("PointsPlaneDistancesWidget/size", self.size())
        settings.setValue("PointsPlaneDistancesWidget/position", self.pos())

        self.sgnWindowClosed.emit()

    def open_help(self):

        dialog = HelpDialog(
            self.plugin_name,
            self.plugin_folder
        )
        dialog.exec_()


class HelpDialog(QDialog):

    def __init__(self,
                 plugin_name,
                 plugin_folder,
                 parent=None):

        super(HelpDialog, self).__init__(parent)

        layout = QVBoxLayout()

        # About section

        helpTextBrwsr = QTextBrowser(self)

        url_path = f"file:///{plugin_folder}/help/help_pts_plane_dists.html"
        helpTextBrwsr.setSource(QUrl(url_path))
        helpTextBrwsr.setSearchPaths([f'{plugin_folder}/help/images/distances'])
        helpTextBrwsr.setMinimumSize(700, 600)

        layout.addWidget(helpTextBrwsr)

        self.setLayout(layout)

        self.setWindowTitle("{} - Points-plane distances help".format(plugin_name))

